# -*- coding: utf-8 -*-
"""Configure logging pour tout automatheque.

Contient quelques loggers par défaut pour les différents modules.
"""

import logging
import logging.config
import json

import yaml

from automatheque.constantes import logger_config_dict

# Python2 n'a pas "FileNotFoundError"
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def configure_logging(conf, desactive_loggers_existants=None):
    """Configure le logging à partir de la conf donnée en paramètre."""
    # TODO https://github.com/yaml/pyyaml/issues/116
    # Tant que pyyaml ne gèrera pas yaml1.2 il faudra faire le test pour savoir
    # si la conf est en json ou en yaml. Ensuite on pourra charger le fichier
    # sans se poser de questions.
    try:
        # logging.config.fileConfig est remplacé par dictConfig, donc on stocke
        # la configuration sous forme de json dans le fichier au lieu du format
        # demandé par fileConfig (ou directement sous forme de dictionnaire).
        if isinstance(conf, str):
            with open(conf, "r") as f:
                try:
                    conf = json.load(f)
                except:
                    pass
                try:
                    conf = yaml.safe_load(conf)
                except:
                    pass
        elif not isinstance(conf, dict):
            raise ValueError("conf doit etre un fichier ou un dictionnaire")
    except FileNotFoundError:
        # configure_logging a déjà été joué au moins une fois avec la config
        # présente dans les constantes, donc recup_logger renvoie bien qqch.
        recup_logger(__name__).warning("Fichier absent : {}".format(conf))
    else:
        if desactive_loggers_existants is not None:
            # on peut forcer True ou False
            conf["disable_existing_loggers"] = desactive_loggers_existants
        logging.config.dictConfig(conf)


def logger_existe(nom):
    """Renvoie True si le logger a déjà été déclaré. False sinon."""
    existe = True
    if nom not in logging.Logger.manager.loggerDict:
        existe = False
    return existe


def recup_logger(name="automatheque"):
    """Renvoie un logger : wrapper autour de logging.

    Charge la configuration de l'application et celle du logging si cela n'a
    pas encore été fait, puis renvoie le logger demandé.
    Renvoie le logger "automatheque" par défaut, défini dans les constantes.

    Par défaut automatheque charge la configuration du fichier définit dans le
    config.ini, dans la section :
    .. code-block:: ini
        [log]
        fichier_config=mon_fichierlog.json

    ou dans le fichier log.json à la racine du script appelant s'il existe.

    :param name:  Nom du logger demandé (passé à logging.getLogger())
    """
    # charge_configuration() utilise aussi recup_logger mais il ne l'appelle
    # que si la configuration n'a pas encore été chargée.
    from automatheque.configuration import charge_configuration

    # Si la configuration a déjà été chargée alors logging est déjà paramétré
    # et charge_configuration() ne fait rien.
    charge_configuration()
    # TODO pour gagner en perf on peut aussi garder un attribut "deja_joue" par
    # exemple pour éviter d'appeler charge_configuration à chaque fois.

    lg = logging.getLogger(name)
    return lg


# D'abord on configure le logging par défaut d'automatheque
configure_logging(logger_config_dict)



# -*- coding: utf-8 -*-
"""Utilitaires de manipulation des fichiers."""


def enleve_caracteres_invalides(value):
    """Supprime les caractères incompatibles avec le système de fichier.

    https://stackoverflow.com/questions/1033424/how-to-remove-bad-path-characters-in-python
    """
    deletechars = '\/:*?"<>|'
    # TODO tester la plateforme
    deletechars = "/"  # Linux autorise quasiment tout le reste
    try:
        for c in deletechars:
            value = value.replace(c, "_")
    except Exception as e:
        # si la valeur est un int ou autre, on la renvoie telle quelle
        pass
    return value



# -*- coding: utf-8 -*-
"""Module pour le patron Fabrique.

Ce patron fournit une interface pour créer des familles d'objets sans spécifier la
classe concrète.  Le patron fabrique, ou méthode fabrique (en anglais factory ou
factory method) est un patron récurrent en génie logiciel.
Une fabrique simple retourne une instance d'une classe parmi plusieurs possibles, en
fonction des paramètres qui ont été fournis. Toutes les classes ont un lien de parenté,
et des méthodes communes, et chacune est optimisée en fonction d'une certaine donnée.

Le patron fabrique abstraite (en anglais abstract factory) va un pas plus loin que la
fabrique simple. Une fabrique abstraite est utilisée pour obtenir un jeu d'objets
connexes.
Par exemple pour implémenter une charte graphique : il existe une fabrique qui retourne
des objets (boutons, menus) dans le style de Windows, une qui retourne des objets dans
le style de Motif, et une dans le style de Macintosh. Une fabrique abstraite est obtenue
en utilisant une fabrique simple.

[wikipedia](https://fr.wikipedia.org/wiki/Fabrique_(patron_de_conception))

cf : https://realpython.com/factory-method-python/

depuis realpython ci dessus :
  The mechanics of Factory Method are always the same. A client
  (SongSerializer.serialize()) depends on a concrete implementation of an interface. It
  requests the implementation from a creator component (get_serializer()) using some
  sort of identifier (format).
"""
from .monteur import Monteur


class Fabrique(object):
    """Classe générique de base pour les "Fabriques".

    Classe pour le design pattern "Factory" pour permettre d'enregistrer les différents
    constructeurs d'objets à la volée et des renvoyer celui qui nous intéresse suivant
    l'objet que l'on souhaite construire.

    On peut
    * soit donner directement une Classe comme constructeur, et on recevra une
    instance de la classe,
    * soit un Monteur concret qui construira l'objet voulu.
     (utile par exemple si l'on veut abstraire le passage de paramètre à l'initialisation
      de la classe)
    """

    def __init__(self):
        self._monteurs = {}

    def enregistre_monteur(self, cle, monteur):
        if not isinstance(monteur, Monteur) and not callable(monteur):
            raise ValueError("Le monteur doit être 'callable'")
        self._monteurs[cle] = monteur
        return monteur

    def recup_monteurs(self):
        return self._monteurs

    def cree(self, cle, *args, **kwargs):
        """Crée l'instance dont le constructeur a la responsabilité.

        Pour les fabriques filles il est conseillé de créer un nouvelle fonction qui
        a une sémantique plus proche de ce qui est créé, et d'appeler la fonction
        Factory.cree dedans, pour rendre le code plus compréhensible.
        """
        monteur = self._monteurs.get(cle)
        if not monteur:
            raise ValueError(f"Pas de monteur {cle}")
        instance = monteur(*args, **kwargs)
        return instance



# -*- coding:utf-8 -*-
"""Utilitaire pour faciliter la création de scripts basés sur automatheque.

TODO cookiecutter : ? https://github.com/LucasRGoes/ports-adapters-sample/blob/master/app/settings.py

Il faut pouvoir l'utiliser aussi dans les modules, pas juste dans les scripts.
TODO !!

Le format que l'on souhaite est le suivant :
```
'''Mon script

Usage:
  mon_script.py [--action] [--config=<fichier_config>]

Options:
  --action  Action realisée
  --config=<fichier_config>  Fichier de configuration [default: mon_script.ini]
'''
__version__ = '0.1a'  # ou importer depuis mon_package.__version__

@script_automatheque(__doc__, __version__)
def main(..., _script=None):
    # code à exécuter
    # _script est un objet `ScriptAutomatheque`
    print(_script)

if __name__ == '__main__':
    main(...)
```
Cela déclenche les logs de début, de fin, parse les arguments etc.
"""

import sys
import os
import re
import datetime
from functools import wraps

import docopt

from automatheque.configuration import charge_configuration, NoOptionError
from automatheque.log import recup_logger, logger_existe


def script_automatheque(chaine_docopt, version=None):
    """Decorateur pour ScriptAutomatheque.

    - signale le début et la fin du traitement
    - passe un objet ScriptAutomatheque à la fonction appelante
    """
    s = ScriptAutomatheque(
        nom=sys.argv[0], chaine_docopt=chaine_docopt, version=version
    )
    s.initialise()

    def decorateur(fonction):
        @wraps(fonction)
        def init_script(*args, **kwds):
            try:
                s.joli_titre()
                s.signale_debut_traitement()
                return fonction(_script=s, *args, **kwds)
            finally:
                s.signale_fin_traitement()

        return init_script

    return decorateur


class ScriptAutomatheque(object):
    """Objet ScriptAutomatheque qui automatise la creation de scripts.

    Cet objet contient :
    - self.arguments : les arguments récupérés par docopt
    - self.config : automatiquement extraite de l'argument --config si présent
                    mergée avec la configuration d'automatheque
    - self.parametres : merge de arguments avec config
    - self.logger : logger pour les appels internes de ScriptAutomatheque

    """

    def __init__(self, nom, chaine_docopt, version=None):
        self.nom = nom
        self.nom_court = re.sub(r"\.py$", "", os.path.basename(nom))
        self.chaine_docopt = chaine_docopt
        self.version = version
        self.arguments = None
        self.config = None
        self._logger = None

    def __repr__(self):
        """TODO utiliser attrs."""
        return "<ScriptAutomatheque {} {} {}>".format(
            self.nom, self.arguments, self.config
        )

    @property
    def logger(self):
        """Loggeur pour le script_automatheque.

        Permet de charger par défaut la configuration du logger stockée dans
        le fichier de configuration, et de logger certaines actions de l'objet
        :class:`ScriptAutomatheque` (alors ces logs sont écrits en utilisant
        la configuration de `nom_du_script.main`)

        L'ideal est ensuite de gérer les loggers avec `logging.getLogger()` ou
        son wrapper `automatheque.log.recup_logger()` même si self.logger est
        quand même accessible dans la fonction wrappée.

        Si le logger pour le script n'a pas été déclaré, alors on utilise le
        logger automatheque.
        """
        if not self._logger:
            if not logger_existe(self.nom_court):
                self._logger = recup_logger()
                self._logger.warning("Aucun logger pour '{}'.".format(self.nom_court))
                self._logger.warning("Utilisation du logger automatheque par défault.")
            else:
                self._logger = recup_logger(self.nom_court)
        return self._logger

    def joli_titre(self):
        self.logger.info("********** {} {} *********".format(self.nom, self.version))

    def initialise(self):
        self.arguments = docopt.docopt(self.chaine_docopt, version=self.version)
        try:
            if self.arguments["--config"]:
                # On écrase les configurations précédentes (par ex si
                # automatheque a déjà été chargé) car on estime que
                # l'annotation vient avec des informations supplémentaires.
                self.config = charge_configuration(
                    [self.arguments["--config"]], ecraser=True, recharger=True
                )
            else:
                self.config = charge_configuration(ecraser=True, recharger=True)
        except KeyError:
            self.config = charge_configuration(ecraser=True, recharger=True)
        # TODO ici on pourrait aussi merger la conf et les arguments ...
        # dans self.parametres ?
        """
        def merge(dict_1, dict_2):
            ""Merge two dictionaries.
            Values that evaluate to true take priority over falsy values.
            `dict_1` takes priority over `dict_2`.
            ""
            return dict((str(key), dict_1.get(key) or dict_2.get(key))
        for key in set(dict_2) | set(dict_1))
        """

    def signale_debut_traitement(self):
        """Signale que le traitement a debuté.

        On pourrait aussi initialiser un logger et s'en servir.
        """
        self.logger.info(
            "{}|{}| Début du traitement".format(
                self.nom, datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%s")
            )
        )

    def signale_fin_traitement(self):
        """Signale que le traitement est terminé."""
        self.logger.info(
            "{}|{}| Fin du traitement".format(
                self.nom, datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%s")
            )
        )



from pathlib import Path

import attr

from automatheque.suivi.ports import StockageAbstraite


@attr.s
class StockageRepertoire(StockageAbstraite):
    repertoire: Path = attr.ib(default=Path("/tmp"))

    def __attr_post_init__(self):
        # TODO >= py3.5
        self.repertoire.mkdir(parents=True, exist_ok=True)

    def existe(self, reference: str):
        # TODO mutex pour thread safety
        fichier = self.repertoire / reference
        return fichier.exists()

    def sauvegarde(self, reference: str, contenu: str):
        fichier = self.repertoire / reference
        with open(fichier, "w") as f:
            f.write(contenu)
        return True



# -*- coding: utf-8 -*-
"""Réglages principaux de Automathèque."""

from os import path

# Répertoire où stocker les principaux réglages de Automathèque.
repertoire_config = "{}/.config/automatheque".format(path.expanduser("~"))

logger_config_dict = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "automatheque": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"}
    },
    "handlers": {
        "automatheque": {
            "level": "INFO",
            "formatter": "automatheque",
            "class": "logging.StreamHandler",
        }
    },
    "loggers": {
        "automatheque": {
            "handlers": ["automatheque"],
            "level": "DEBUG",
            "propagate": True,
        }
    },
}



# -*- coding: utf-8 -*-
"""Module pour le patron Monteur.

Ce patron sépare le processus de construction d'un objet de sa représentation. Cela
permet d'utiliser le même processus pour obtenir différents résultats.

Dans les faits une classe délègue la création de ses instances à un Monteur, mais garde
le contrôle sur le processus de création.

Ainsi il faut créer des Monteurs "concrets" (sous-classe) pour chaque représentation
différente, qui respectent l'interface de montage de la classe exigée.

Ce patron est particulièrement utile quand il y a de nombreux paramètres de création,
presque tous optionnels.
"""


class Monteur(object):
    """Classe abstraite pour le patron Monteur.

    Cela permet d'avoir plusieurs filles du constructeur différentes pour générer les
    instances différemment ou renvoyer des instances de classes différentes étant
    donné les arguments en entrée du monteur.

    Le processus de création est par défaut défini dans :Monteur.construit:
    """

    CLASSE = None

    def __call__(self, *args, **kwargs):
        """Facilite la construction des objets."""
        return self.construit(*args, **kwargs)

    def construit(self, *args, **kwargs):
        """Fonction à surcharger par les Monteurs concrets.

        Par défaut cela renvoie juste une instance de la classe self.CLASSE.
        """
        if self.CLASSE is None:
            raise NotImplementedError
        return self.CLASSE()



from __future__ import annotations
from typing import TYPE_CHECKING, Union

from automatheque.conception.structures import MetaInstancePersistanteRegistre

if TYPE_CHECKING:
    from automatheque.greffon.greffon import Greffon


class RegistreGreffons(metaclass=MetaInstancePersistanteRegistre):
    """
    Stocke toutes les instances des greffons qui héritent de `~automatheque.greffon.Greffon`.

    Fournit plusieurs méthodes de classe pour récupérer les Greffons par nom ou par capacité.
    """

    @classmethod
    def greffons_par_capacite(cls, capacite: str):
        return [
            p for p in cls._instances(inclure_enfants=True) if capacite in p.capacites
        ]

    @classmethod
    def greffons_identifiants(cls):
        return [p.identifiant for p in cls._instances(inclure_enfants=True)]

    @classmethod
    def greffon_par_identifiant(cls, identifiant: str) -> Union[Greffon, None]:
        greffons = [
            p
            for p in cls._instances(inclure_enfants=True)
            if p.identifiant == identifiant
        ]
        return greffons[0] if greffons else None



# -*- coding: utf-8 -*-
"""Chargement du fichier de configuration général."""

from os import path

from configparser import ConfigParser, NoOptionError, NoSectionError

from automatheque import constantes
from automatheque.log import recup_logger, configure_logging

fichier_config = "{}/config.ini".format(constantes.repertoire_config)

# Pour pouvoir tester relativement a une erreur automatheque:
NoOptionError = NoOptionError
NoSectionError = NoSectionError


def charge_configuration(fichiers_supplementaires=[], ecraser=False, recharger=False):
    """Fonction pour charger la configuration generale de automatheque."""
    if hasattr(charge_configuration, "config") and not recharger:
        return charge_configuration.config

    charge_configuration.config = ConfigParser()

    # On charge le logger ici, donc si la configuration a déjà été chargée on
    # est sûrs que recup_logger ne sera pas appelé de nouveau.
    # (ce que l'on souhaite vu que recup_logger appelle charge_configuration..)
    logger = recup_logger(__name__)

    # Pour écraser la configuration de automatheque on la charge en premier
    # et les fichiers suivants vont prendre la précédence.
    if ecraser:
        _charge_fichier_configuration(fichier_config, charge_configuration.config)

    # Ajout des autres fichiers s'il y en a :
    for f in fichiers_supplementaires:
        # Si "f" n'est pas un fichier, on regarde s'il existe dans le
        # répertoire de configuration d'automatheque.
        paths_a_tester = [f, path.join(constantes.repertoire_config, f)]
        for fichier in paths_a_tester:
            if not path.isfile(fichier):
                logger.debug("{} n'est pas un fichier.".format(fichier))
                continue

            _charge_fichier_configuration(fichier, charge_configuration.config)
            break  # TODO sauf si on veut charger les 2 fichiers ?
    # Si on veut conserver la configuration de automatheque, alors on charge
    # sa configuration en dernier :
    if not ecraser:
        _charge_fichier_configuration(fichier_config, charge_configuration.config)

    # Si la configuration que l'on vient d'importer contient des paramètres
    # qui concernent les logs alors on configure le logging :
    _configure_logging(charge_configuration.config)

    return charge_configuration.config


def _charge_fichier_configuration(fichier, config):
    recup_logger(__name__).debug(
        "Chargement du fichier de configuration : {}.".format(fichier)
    )
    config.read(fichier)


def _recup_fichier_config_log(config):
    """Renvoie le fichier de configuration des logs.

    S'il n'en trouve pas dans la configuration globale il renvoie log.json par
    défaut.
    """
    try:
        # TODO on ne doit pas forcer l'existence de la config 'log'
        fichier_config_log = config.get("log", "fichier_config", fallback="log.json")
    except NoOptionError:
        # TODO créer une exception custom
        msg = """Pas de configuration pour le fichier de log.
        Rappel structure:
        [log]
        fichier_config=chemin/vers/fichier/config.json ou .yaml"""
        recup_logger(__name__).exception(msg)
        raise
    return fichier_config_log


def _configure_logging(config):
    """Configure logging avec les valeurs définies dans le config.ini.

    Par défaut automatheque charge la configuration du fichier définit dans le
    config.ini dans la section :
    `[log]
    fichier_config=mon_fichierlog.json ou yaml
    `
    ou dans le fichier log.json à la racine du script s'il existe.

    Cette fonction est appelée automatiquement par charge_configuration à sa
    première exécution.

    :param desactive_loggers_existants: boolean Force ce paramètre de log
    """
    # Puis on ajoute la configuration trouvée dans le fichier de configuration
    # du logging, dont le chemin est stocké dans la configuration globale :

    try:
        fichier_config_log = _recup_fichier_config_log(config)
        desactive_loggers_existants = config.get(
            "log", "desactive_loggers_existants", fallback=None
        )
    except Exception:
        pass
    else:
        configure_logging(fichier_config_log, desactive_loggers_existants)



# -*- coding: utf-8 -*-
"""Module pour le patron Adaptateur.

Ce patron convertit l'interface d'une classe en une autre interface exploitée ailleurs.
 
Cela permet d'interconnecter des classes qui sans cela seraient incompatibles. Il est 
utilisé dans le cas où un programme se sert d'une bibliothèque de classe qui ne 
correspond plus à l'utilisation qui en est faite, à la suite d'une mise à jour de la 
bibliothèque dont l'interface a changé. 

Un objet adaptateur (en anglais adapter) expose alors l'ancienne interface en utilisant
les fonctionnalités de la nouvelle.

[wikipedia](https://fr.wikipedia.org/wiki/Adaptateur_(patron_de_conception))

source : https://github.com/faif/python-patterns/blob/master/patterns/structural/adapter.py
"""


class Adaptateur(object):
    """Remplace les méthodes d'un objet par des méthodes d'un autre objet.

    :usage:
    objet = Objet()
    objet = Adaptateur(objet, nouvelle_methode=autre_objet.methode_origine)
    objet.nouvelle_methode()  # exécute autre_objet.methode_origine()
    """

    def __init__(self, obj, **nouvelles_methodes):
        """Les méthodes sont ajoutées au dict de l'instance d'Adaptateur."""
        self.obj = obj
        self.__dict__.update(nouvelles_methodes)

    def __getattr__(self, attr):
        """Tous les appels non "adaptés" sont transmis à l'objet sous-jacent."""
        return getattr(self.obj, attr)

    def dict_originel(self):
        """Renvoie le dict originel de l'objet sous-jacent."""
        return self.obj.__dict__



import attr

from automatheque.suivi.adaptateurs.repertoire import StockageRepertoire
from automatheque.suivi.ports import StockageAbstraite


@attr.s
class JournalSuivi:
    stockage: StockageAbstraite = attr.ib(default=StockageRepertoire())  # handler ?

    def deja_fait(self, reference: str) -> bool:
        return self.stockage.existe(reference)

    # decorator !!
    def ni_fait_ni_à_refaire(self, nom_arg_reference: str):
        """Décore une fonction pour qu'elle ne soit jouée qu'une fois pour un argument donné.

        Il faut donner en entrée le nom du paramètre qui contient la référence, et la
        decoration sauvegardera "nom_fonction_reference".

        TODO ajouter option pour date.
        """
        pass

    def coche(self, reference: str, contenu: str) -> bool:
        return self.stockage.sauvegarde(reference, contenu)
